-------------------------------------------------------------------------
-- AtoneWatch v8.0.1.6           A Visual Atonement Buff Monitor for WoW.
-- Copyright (C) 2018  Elshandra (elshandra[at]forkatz[dot]com)
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <https://www.gnu.org/licenses/>.
-------------------------------------------------------------------------
local aw = _G.AtoneWatch

local awCP = {}
aw.CP = awCP

-- Build the Configuration Frame
function awCP:BuildFrame()
  local f = CreateFrame("Frame", "AWConfigFrame", UIParent)
  awCP.Frame = f
  f.name = "AtoneWatch"
  -- Enabled Checkbox
  f.btnEnabled = awCP:CreateCheckBox("Enabled", "Enable AtoneWatch", "Enable or Disable AtoneWatch", awCP.UpdateEnabled)
  f.btnEnabled:SetPoint("TOPLEFT", 10, -10)
  -- Show Checkbox
  f.btnHideUI = awCP:CreateCheckBox("Show", "Show AtoneWatch", "Show or Hide AtoneWatch", awCP.UpdateHideUI)
  f.btnHideUI:SetPoint("TOPLEFT", f.btnEnabled, "BOTTOMLEFT", 0, -5)
  -- Show Title Checkbox
  f.btnUIHideTitle = awCP:CreateCheckBox("UIHideTitle", "Show Title", "Show or Hide the frame title", awCP.UpdateUIHideTitle)
  f.btnUIHideTitle:SetPoint("TOPLEFT", f.btnHideUI, "BOTTOMLEFT", 0, -5)
  -- Lock Frame Checkbox
  f.btnLock = awCP:CreateCheckBox("Lock", "Lock Frame", "Lock the atonements fame", awCP.UpdateLock)
  f.btnLock:SetPoint("LEFT", f.btnHideUI.Text, "RIGHT", 15, 0)
  -- Grow Direction
  ---- Label
  f.lblGrowDir = f:CreateFontString(nil, "OVERLAY", "GameFontWhite")
  f.lblGrowDir:SetText("Grow Direction:")
  f.lblGrowDir:SetPoint("TOPLEFT", f.btnUIHideTitle, "BOTTOMLEFT", 5, -8)
  ---- Up
  f.btnGrowUp = awCP:CreateCheckBox("GrowUp", "Up", "Bars will grow upwards", awCP.UpdateGrowUp)
  f.btnGrowUp:SetPoint("LEFT", f.lblGrowDir, "RIGHT", 5, 0)
  ---- Down
  f.btnGrowDown = awCP:CreateCheckBox("GrowDown", "Down", "Bars will grow downwards", awCP.UpdateGrowDown)
  f.btnGrowDown:SetPoint("LEFT", f.btnGrowUp.Text, "RIGHT", 5, 0)
  InterfaceOptions_AddCategory(aw.CP.Frame)
  -- Sort Order
  ---- Label
  f.lblSortDir = f:CreateFontString(nil, "OVERLAY", "GameFontWhite")
  f.lblSortDir:SetText("Sort Order:")
  f.lblSortDir:SetPoint("TOPLEFT", f.lblGrowDir, "BOTTOMLEFT", 0, -17)
  ---- Descending
  f.btnSortDesc = awCP:CreateCheckBox("SortDesc", "Descending", "Shortest duration will be at the top", awCP.UpdateSortDesc)
  f.btnSortDesc:SetPoint("TOPLEFT", f.btnGrowUp, "BOTTOMLEFT", 0, -5)
  ---- Ascending
  f.btnSortAsc = awCP:CreateCheckBox("SortAsc", "Ascending", "Shortest duration will be at the bottom", awCP.UpdateSortAsc)
  f.btnSortAsc:SetPoint("LEFT", f.btnSortDesc.Text, "RIGHT", 5, 0)
  -- Linger Time
  ---- Label
  f.lblLinger = f:CreateFontString(nil, "OVERLAY", "GameFontWhite")
  f.lblLinger:SetText("Linger Duration:")
  f.lblLinger:SetPoint("TOPLEFT", f.lblSortDir, "BOTTOMLEFT", 0, -27)
  ---- Slider
  f.slLinger = CreateFrame("Slider", "AWCPLinger", f, "OptionsSliderTemplate")
  f.slLinger:SetPoint("LEFT", f.lblLinger, "RIGHT", 10, 0)
  f.slLinger:SetWidth(300)
  f.slLinger:SetHeight(20)
  f.slLinger:SetOrientation('HORIZONTAL')
  f.slLinger:SetMinMaxValues(0, 10)
  f.slLinger:SetValueStep(0.1)
  f.slLinger:SetObeyStepOnDrag(true)
  f.slLinger.tooltipText = "Set the duration for the bar to remain after the buff expires (seconds)"
  getglobal(f.slLinger:GetName() .. 'Low'):SetText('0')
  getglobal(f.slLinger:GetName() .. 'High'):SetText('10')
  f.slLinger:SetScript("OnValueChanged", function(self,value) awCP:UpdateLinger(self, value) end)
  -- UI Width
  ---- Label
  f.lblUIWidth = f:CreateFontString(nil, "OVERLAY", "GameFontWhite")
  f.lblUIWidth:SetText("Frame Width:")
  f.lblUIWidth:SetPoint("TOPLEFT", f.lblLinger, "BOTTOMLEFT", 0, -27)
  ---- Slider
  f.slUIWidth = CreateFrame("Slider", "AWCPUIWidth", f, "OptionsSliderTemplate")
  f.slUIWidth:SetPoint("TOPLEFT", f.slLinger, "BOTTOMLEFT", 0, -20)
  f.slUIWidth:SetWidth(300)
  f.slUIWidth:SetHeight(20)
  f.slUIWidth:SetOrientation('HORIZONTAL')
  f.slUIWidth:SetMinMaxValues(100, 300)
  f.slUIWidth:SetValueStep(5)
  f.slUIWidth:SetObeyStepOnDrag(true)
  f.slUIWidth.tooltipText = "Set the width of the Atonements frame"
  getglobal(f.slUIWidth:GetName() .. 'Low'):SetText('100')
  getglobal(f.slUIWidth:GetName() .. 'High'):SetText('300')
  f.slUIWidth:SetScript("OnValueChanged", function(self,value) awCP:UpdateUIWidth(self, value) end)
  -- Bar Height
  ---- Label
  f.lblUIBarHeight = f:CreateFontString(nil, "OVERLAY", "GameFontWhite")
  f.lblUIBarHeight:SetText("Bar Height:")
  f.lblUIBarHeight:SetPoint("TOPLEFT", f.lblUIWidth, "BOTTOMLEFT", 0, -28)
  ---- Slider
  f.slUIBarHeight = CreateFrame("Slider", "AWCPUIBarHeight", f, "OptionsSliderTemplate")
  f.slUIBarHeight:SetPoint("TOPLEFT", f.slUIWidth, "BOTTOMLEFT", 0, -20)
  f.slUIBarHeight:SetWidth(300)
  f.slUIBarHeight:SetHeight(20)
  f.slUIBarHeight:SetOrientation('HORIZONTAL')
  f.slUIBarHeight:SetMinMaxValues(10, 50)
  f.slUIBarHeight:SetValueStep(1)
  f.slUIBarHeight:SetObeyStepOnDrag(true)
  f.slUIBarHeight.tooltipText = "Set the individual bar height"
  getglobal(f.slUIBarHeight:GetName() .. 'Low'):SetText('10')
  getglobal(f.slUIBarHeight:GetName() .. 'High'):SetText('50')
  f.slUIBarHeight:SetScript("OnValueChanged", function(self,value) awCP:UpdateUIBarHeight(self, value) end)
  -- Scale
  ---- Label
  f.lblScale = f:CreateFontString(nil, "OVERLAY", "GameFontWhite")
  f.lblScale:SetText("Scale:")
  f.lblScale:SetPoint("TOPLEFT", f.lblUIBarHeight, "BOTTOMLEFT", 0, -28)
  ---- Slider
  f.slScale = CreateFrame("Slider", "AWCPScale", f, "OptionsSliderTemplate")
  f.slScale:SetPoint("TOPLEFT", f.slUIBarHeight, "BOTTOMLEFT", 0, -20)
  f.slScale:SetWidth(300)
  f.slScale:SetHeight(20)
  f.slScale:SetOrientation('HORIZONTAL')
  f.slScale:SetMinMaxValues(0.1, 2)
  f.slScale:SetValueStep(0.05)
  f.slScale:SetObeyStepOnDrag(true)
  f.slScale.tooltipText = "Set the Scale of the Atonements frame"
  getglobal(f.slScale:GetName() .. 'Low'):SetText('0.1')
  getglobal(f.slScale:GetName() .. 'High'):SetText('2')
  f.slScale:SetScript("OnValueChanged", function(self,value) awCP:UpdateScale(self, value) end)
  -- Initialise the objects.
  awCP:ReadConfiguration()
  f:Hide()
  -- Add Demo Bars Button
  f.btnAddBar = CreateFrame("Button", "AWCPAddBar", f, "UIPanelButtonTemplate")
  f.btnAddBar:SetSize(80, 22)
  f.btnAddBar:SetPoint("TOPRIGHT", f, "TOPRIGHT", -20, -10)
  f.btnAddBar:SetScript("OnClick", awCP.AddBar)
  f.btnAddBar:SetText("Add Demo Bar")
  -- Version Label
  f.lblVersion = f:CreateFontString(nil, "OVERLAY", "GameFontHighlightLarge")
  f.lblVersion:SetTextColor(1, 1, 0.5, 0.9)
  f.lblVersion:SetText("AtoneWatch v" .. tostring(aw.Data.Version))
  f.lblVersion:SetPoint("BOTTOMRIGHT", f, "BOTTOMRIGHT", -5, 5)
  -- OnShow Event
  f:SetScript("OnShow", awCP.OnShow)
end

-- Frame handler to show tooltips over checkboxes
function awCP:FrameOnEnter(self)
  if self.tooltip == nil or self.tooltip == "" then return end
  GameTooltip:Hide()
  GameTooltip:SetOwner(self, "ANCHOR_TOPLEFT")
  GameTooltip:ClearLines()
  GameTooltip:SetText(self.tooltip)
  GameTooltip:Show()
end

-- Frame handler to hide tooltips over checkboxes
function awCP:FrameOnLeave(self)
  if self.tooltip == nil or self.tooltip == "" then return end
  GameTooltip:Hide()
end

-- Enabled Checkbox Click Event
function awCP:UpdateEnabled(button, down)
  local cb = awCP.Frame.btnEnabled
  if (cb.IsChecked) then
    cb.IsChecked = false
    cb.CheckMark:Hide()
    aw.ui:Hide()
    awConfig.Enabled = false
  else
    cb.IsChecked = true
    cb.CheckMark:Show()
    awConfig.Enabled = true
    if not awConfig.HideUI then aw.ui:Show() end
  end
end

-- HideUI Checkbox Click Event
function awCP:UpdateHideUI(button, down)
  local cb = awCP.Frame.btnHideUI
  if (cb.IsChecked) then
    cb.IsChecked = false
    cb.CheckMark:Hide()
    awConfig.HideUI = true
    aw.ui:Hide()
  else
    cb.IsChecked = true
    cb.CheckMark:Show()
    awConfig.HideUI = false
    aw.ui:Show()
  end
end

-- Lock Checkbox Click Event
function awCP:UpdateLock(button, down)
  local cb = awCP.Frame.btnLock
  if cb.IsChecked then
    cb.IsChecked = false
    cb.CheckMark:Hide()
    if not awConfig.UIHideTitle then
      aw.Frame:SetScript("OnDragStart", aw.Frame.StartMoving)
    end
  else
    cb.IsChecked = true
    cb.CheckMark:Show()
    awConfig.LockUI = true
    aw.Frame:SetScript("OnDragStart", nil)
  end
end

-- GrowUp Checkbox Click Event
function awCP:UpdateGrowUp(button, down)
  local cb = awCP.Frame.btnGrowUp
  local cbd = awCP.Frame.btnGrowDown
  if (cb.IsChecked) then return end
  -- Change the tooltip on the sort order checkboxes.
  awCP.Frame.btnSortAsc.tooltip = "Shortest duration will be at the bottom"
  awCP.Frame.btnSortDesc.tooltip = "Shortest duration will be at the top"
  -- Set the Radio Button states.
  cbd.CheckMark:Hide()
  cbd.IsChecked = false
  cb.IsChecked = true
  cb.CheckMark:Show()
  -- Change Grow Direction
  if awConfig.GrowDown then
    awConfig.GrowDown = false
    aw.ui:ReverseGrowth()
  end
end

-- GrowDown Checkbox Click Event
function awCP:UpdateGrowDown(button, down)
  local cb = awCP.Frame.btnGrowDown
  local cbu = awCP.Frame.btnGrowUp
  if (cb.IsChecked) then return end
  -- Change the tooltip on the sort order checkboxes.
  awCP.Frame.btnSortAsc.tooltip = "Shortest duration will be at the top"
  awCP.Frame.btnSortDesc.tooltip = "Shortest duration will be at the bottom"
  -- Set the Radio Button states.
  cbu.CheckMark:Hide()
  cbu.IsChecked = false
  cb.IsChecked = true
  cb.CheckMark:Show()
  -- Change Grow Direction
  if not awConfig.GrowDown then
    awConfig.GrowDown = true
    aw.ui:ReverseGrowth()
  end
end

-- UIHideTitle Checkbox Click Event
function awCP:UpdateUIHideTitle(button, down)
  local cb = awCP.Frame.btnUIHideTitle
  if cb.IsChecked then
    cb.IsChecked = false
    cb.CheckMark:Hide()
    if awConfig.UIHideTitle then return end
    aw.ui:HideTitle()
  else
    cb.IsChecked = true
    cb.CheckMark:Show()
    if awConfig.UIHideTitle then
      aw.ui:ShowTitle()
    end
  end
end

-- SortDesc Checkbox Click Event
function awCP:UpdateSortDesc(button, down)
  local cb = awCP.Frame.btnSortDesc
  local cba = awCP.Frame.btnSortAsc
  if cb.IsChecked then return end
  cb.IsChecked = true
  cb.CheckMark:Show()
  cba.IsChecked = false
  cba.CheckMark:Hide()
  -- Change the Sort Order
  if awConfig.SortAsc then
    awConfig.SortAsc = false
    aw.ui:RefreshUI()
  end
end

-- SortAsc Checkbox Click Event
function awCP:UpdateSortAsc(button, down)
  local cb = awCP.Frame.btnSortAsc
  local cbd = awCP.Frame.btnSortDesc
  if cb.IsChecked then return end
  cb.IsChecked = true
  cb.CheckMark:Show()
  cbd.IsChecked = false
  cbd.CheckMark:Hide()
  -- Change the Sort Order
  if awConfig.SortAsc then return end
  awConfig.SortAsc = true
  aw.ui:RefreshUI()
end

-- Linger Slider Value Changed Event
function awCP:UpdateLinger(slider, value)
  awConfig.Linger = tonumber(format('%.1f', value))
  getglobal(slider:GetName() .. 'Text'):SetText(format('%.1f', value))
end

-- UIWidth Slider Value Changed Event
function awCP:UpdateUIWidth(slider, value)
  awConfig.UIWidth = tonumber(format('%d', value))
  getglobal(slider:GetName() .. 'Text'):SetText(format('%d', value))
  aw.ui:UpdateWidth()
end

-- UIBarHeight Slider Value Changed Event
function awCP:UpdateUIBarHeight(slider, value)
  awConfig.UIBarHeight = tonumber(format('%d', value))
  getglobal(slider:GetName() .. 'Text'):SetText(format('%d', value))
  aw.ui:UpdateHeight()
end

-- Scale Slider Value Changed Event
function awCP:UpdateScale(slider, value)
  awConfig.Scale = tonumber(format('%.2f', value))
  getglobal(slider:GetName() .. 'Text'):SetText(format('%.2f', value))
  aw.ui:Scale(tonumber(format('%.2f', value)))
end

-- Create a Checkbox Frame that doesn't have a retarded mouse region.
function awCP:CreateCheckBox(sName, sText, sToolTip, fOnClick)
  -- Checkbox Frame
  local f = CreateFrame("Button", "AWCP" .. sName, awCP.Frame)
  f:ClearAllPoints()
  f:SetHeight(24)
  f:SetWidth(24)
  f:SetBackdrop({
    bgFile = "Interface\\BUTTONS\\UI-Checkbox-Up",
    edgefile = nil, edgesize = 0,
    insets = { left = 0, right = 0, top = 0, bottom = 0, },
  })
  f:SetHighlightTexture("Interface\\BUTTONS\\ButtonHilight-Round", "ADD")
  f:SetNormalFontObject("GameFontWhite")
  -- CheckMark Texture
  f.CheckMark = f:CreateTexture(nil, "OVERLAY")
  f.CheckMark:SetTexture("Interface\\BUTTONS\\UI-Checkbox-Check")
  f.CheckMark:SetPoint("CENTER")
  f.CheckMark:SetHeight(20)
  f.CheckMark:SetWidth(20)
  f.CheckMark:Hide()
  -- Label Frame
  f.Text = f:CreateFontString(nil, "OVERLAY", "GameFontWhite")
  f.Text:SetText(sText)
  f.Text:SetPoint("LEFT", f, "RIGHT", 2, 0)
  -- Data Storage
  f.IsChecked = false
  f.tooltip = sToolTip
  -- Script Handlers for Click & Tooltip
  f:SetScript("OnClick", fOnClick)
  f:SetScript("OnEnter", function(self) awCP:FrameOnEnter(self) end)
  f:SetScript("OnLeave", function(self) awCP:FrameOnLeave(self) end)
  return f
end

-- AddBar Button Click Event
function awCP:AddBar()
  i = #aw.Data.Units
  while aw:DataGetUnit("FapFapFap" .. tostring(i)) ~= nil do
    i = i + 1
    if i >= 30 then return end
  end
  aw:DataAddUnit("FapFapFap" .. tostring(i), "Nonsense Player " .. tostring(#aw.Data.Units), GetTime() + 15)
  local UnitIndex = aw:DataGetUnit("FapFapFap" .. tostring(i))
  aw.Data.Units[UnitIndex].UnitClass = select(2, GetClassInfo(random(1, GetNumClasses())))
end

-- OnShow Event
function awCP:OnShow()
  awCP:ReadConfiguration()
end

-- Read the Configuration
function awCP:ReadConfiguration()
  local f = aw.CP.Frame
  -- Enabled
  if awConfig.Enabled then
    f.btnEnabled.IsChecked = true
    f.btnEnabled.CheckMark:Show()
  else
    f.btnEnabled.IsChecked = false
    f.btnEnabled.CheckMark:Hide()
  end
  -- HideUI (Inversed)
  if awConfig.HideUI then
    f.btnHideUI.IsChecked = false
    f.btnEnabled.CheckMark:Hide()
  else
    f.btnHideUI.IsChecked = true
    f.btnHideUI.CheckMark:Show()
  end
  -- Grow Direction
  if awConfig.GrowDown then
    f.btnGrowDown.IsChecked = true
    f.btnGrowDown.CheckMark:Show()
    f.btnGrowUp.IsChecked = false
    f.btnGrowUp.CheckMark:Hide()
    f.btnSortAsc.tooltip = "Shortest duration will be at the bottom"
    f.btnSortDesc.tooltip = "Shortest duration will be at the top"
  else
    f.btnGrowDown.IsChecked = false
    f.btnGrowDown.CheckMark:Hide()
    f.btnGrowUp.IsChecked = true
    f.btnGrowUp.CheckMark:Show()
    f.btnSortAsc.tooltip = "Shortest duration will be at the bottom"
    f.btnSortDesc.tooltip = "Shortest duration will be at the top"
  end
  -- UIHideTitle (Inversed)
  if awConfig.UIHideTitle then
    f.btnUIHideTitle.IsChecked = false
    f.btnUIHideTitle.CheckMark:Hide()
  else
    f.btnUIHideTitle.IsChecked = true
    f.btnUIHideTitle.CheckMark:Show()
  end
  -- Sort Order
  if awConfig.SortAsc then
    f.btnSortAsc.IsChecked = true
    f.btnSortAsc.CheckMark:Show()
    f.btnSortDesc.IsChecked = false
    f.btnSortDesc.CheckMark:Hide()
  else
    f.btnSortAsc.IsChecked = false
    f.btnSortAsc.CheckMark:Hide()
    f.btnSortDesc.IsChecked = true
    f.btnSortDesc.CheckMark:Show()
  end
  -- Linger
  if awConfig.Linger then
    f.slLinger:SetValue(awConfig.Linger)
    getglobal(f.slLinger:GetName() .. 'Text'):SetText(tostring(awConfig.Linger))
  end
  -- UIWidth
  if awConfig.UIWidth then
    f.slUIWidth:SetValue(awConfig.UIWidth)
    getglobal(f.slUIWidth:GetName() .. 'Text'):SetText(tostring(awConfig.UIWidth))
  end
  -- Bar Height
  if awConfig.UIBarHeight then
    f.slUIBarHeight:SetValue(awConfig.UIBarHeight)
    getglobal(f.slUIBarHeight:GetName() .. 'Text'):SetText(tostring(awConfig.UIBarHeight))
  end
  -- Scale
  if awConfig.Scale then
    f.slScale:SetValue(awConfig.Scale)
    getglobal(f.slScale:GetName() .. 'Text'):SetText(tostring(awConfig.Scale))
  end
end
