------------------------------------------------------------------------
-- AtoneWatch v8.0.1.6          A Visual Atonement Buff Monitor for WoW.
-- Copyright (C) 2018  Elshandra (elshandra[at]forkatz[dot]com)
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <https://www.gnu.org/licenses/>.
-------------------------------------------------------------------------
-- TODO:
--   01) Calculate proper atonement durations instead of using static values.
--   02) Consider azerite traits such as Enduring Luminescence, 
--       Moment of Repose, etc.
--   03) Set font/size.
-------------------------------------------------------------------------
_G.AtoneWatch = {}

local aw = _G.AtoneWatch

-- Spell ID Reference
local awSpellIDs = {
  Atonement = 194384,
  Evangelism = 246287,
  Radiance = 194509,
  Shield = 17,
  ShadowMend = 186263
}

-- Spell Durations
-- TODO-01: Figure out how to calculate/measure these properly ;o)
local awDurations = {
  Evangelism = 6,
  Radiance = 9,
  Shield = 15,
  ShadowMend = 15
}

-- Data Table
-- Player { Name, GUID, Spec }
-- Units { UnitName, UnitGUID, BuffApplied, BuffExpiry, UnitUnitID, HealthPct, BuffRem, UnitClass }
-- LastSpellID
-- NextDuration -- same as awDuarations[LastSpellID] 
-- LastUpdate
-- MaxDuration
aw.Data = {
  Player={},
  Units={},
  LastSpellID=0,
  NextDuration=15,
  LastUpdate=0,
  UILoaded = false,
  MaxDuration=0,
  Version = "8.0.1.6"
}

if awConfig == nil then
  awConfig = {}
  awConfig.Enabled = true
  awConfig.Linger = 2
  awConfig.Debug = false
  awConfig.GrowDown = false
  awConfig.SortAsc = false
  awConfig.UIWidth = 200
  awConfig.UIBarHeight = 18
  awConfig.UIHideTitle = false
end

aw.Frame = CreateFrame("frame", "AtoneWatch", UIParent)
aw.Frame:RegisterEvent("PLAYER_LOGIN")
aw.Frame:RegisterEvent("ADDON_LOADED")
aw.Frame:RegisterEvent("PLAYER_ENTERING_WORLD")
aw.Frame:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED")
aw.Frame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")

-------------------
-- Slash Commands -
-------------------
SLASH_ATONEWATCH1 = '/aw'
SlashCmdList["ATONEWATCH"] = function(msg)
  local cmd, args = msg:match("^(%S*)%s*(.-)$")
  cmd = strlower(cmd)
  if cmd == "" or cmd == "help" or cmd == "?" then
    aw:SlashHelp()
  elseif cmd == "on"     then aw:SlashOn()
  elseif cmd == "off"    then aw:SlashOff()
  elseif cmd == "show"   then aw:SlashShow(args)
  elseif cmd == "hide"   then aw:SlashHide(args)
  elseif cmd == "grow"   then aw:SlashGrow(args)
  elseif cmd == "lock"   then aw:SlashLock(args)
  elseif cmd == "linger" then aw:SlashLinger(args)
  elseif cmd == "scale"  then aw:SlashScale(args)
  elseif cmd == "sort"   then aw:SlashSort(args)
  elseif cmd == "width"  then aw:SlashWidth(args)
  elseif cmd == "debug"  then aw:SlashDebug(args)
  elseif cmd == "height" then aw:SlashHeight(args)
  elseif cmd == "config" then aw:SlashConfig()
  else
    DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch: Unrecognised /aw command.", 1, 1, 1)
  end
end

-- /aw Help
function aw:SlashHelp()
  DEFAULT_CHAT_FRAME:AddMessage("Atone Watch Slash Commands", 1, 1, 0.45)
  DEFAULT_CHAT_FRAME:AddMessage("----------------------------------", 1, 1, 0.45)
  DEFAULT_CHAT_FRAME:AddMessage("  /aw on|off - Enable/Disable AtoneWatch", 1, 1, 1)
  DEFAULT_CHAT_FRAME:AddMessage("  /aw show|hide [title] - Show or Hide the Atonements window/title", 1, 1, 1)
  DEFAULT_CHAT_FRAME:AddMessage("  /aw grow up|down - Set the grow direction for bars.", 1, 1, 1)
  DEFAULT_CHAT_FRAME:AddMessage("  /aw sort asc|desc - Set the expiration sort order", 1, 1, 1)
  DEFAULT_CHAT_FRAME:AddMessage("  /aw lock - Lock the Atonements window in place", 1, 1, 1)
  DEFAULT_CHAT_FRAME:AddMessage("  /aw width # - Set the width to #", 1, 1, 1)
  DEFAULT_CHAT_FRAME:AddMessage("  /aw height # - Set the bar height to #", 1, 1, 1)
  DEFAULT_CHAT_FRAME:AddMessage("  /aw linger # - Keep for # seconds after expiry", 1, 1, 1)
end

-- /aw config
function aw:SlashConfig(args)
  if args and args == "show" then
    if awConfig.Enabled then 
      DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch - Current Configuration (Enabled)", 1, 1, 0.45)
    else
      DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch - Current Configuration (Disabled)", 1, 1, 0.45)
    end
    if awConfig.Linger ~= nil then
      DEFAULT_CHAT_FRAME:AddMessage("Linger Time: " .. tostring(awConfig.Linger) .. " seconds", 1, 1, 0.45)
    else
      DEFAULT_CHAT_FRAME:AddMessage("Linger Time: 2 seconds", 1, 1, 0.45)
    end
    if awConfig.GrowDown == nil or awConfig.GrowDown == false then
      DEFAULT_CHAT_FRAME:AddMessage("Grow Direction: Upward", 1, 1, 0.45)
    else
      DEFAULT_CHAT_FRAME:AddMessage("Grow Direction: Downward", 1, 1, 0.45)
    end
    if awConfig.SortAsc == nil or awConfig.SortAsc == false then
      DEFAULT_CHAT_FRAME:AddMessage("Sort Order: Descending", 1, 1, 0.45)
    else
      DEFAULT_CHAT_FRAME:AddMessage("Sort Order: Acending", 1, 1, 0.45)
    end
    if awConfig.Scale == nil then 
      DEFAULT_CHAT_FRAME:AddMessage("Frame Scale: 1", 1, 1, 0.45)
    else
      DEFAULT_CHAT_FRAME:AddMessage("Frame Scale: " .. tostring(awConfig.Scale), 1, 1, 0.45)
    end
    if awConfig.LockUI == nil or awConfig.LockUI == false then
      DEFAULT_CHAT_FRAME:AddMessage("Frame Lock: Disabled", 1, 1, 0.45)
    else
      DEFAULT_CHAT_FRAME:AddMessage("Frame Lock: Enabled", 1, 1, 0.45)
    end
    if awConfig.UIHideTitle == nil or awConfig.UIHideTitle == false then
      DEFAULT_CHAT_FRAME:AddMessage("Frame Title: On", 1, 1, 0.45)
    else
      DEFAULT_CHAT_FRAME:AddMessage("Frame Title: Off", 1, 1, 0.45)
    end
    if awConfig.HideUI == nil or awConfig.HideUI == false then
      DEFAULT_CHAT_FRAME:AddMessage("Frame Visible: Yes", 1, 1, 0.45)
    else
      DEFAULT_CHAT_FRAME:AddMessage("Frame Visible: No", 1, 1, 0.45)
    end
    if awConfig.UIWidth == nil then
      DEFAULT_CHAT_FRAME:AddMessage("Frame Width: 200", 1, 1, 0.45)
    else
      DEFAULT_CHAT_FRAME:AddMessage("Frame Width: " .. tostring(awConfig.UIWidth), 1, 1, 0.45)
    end
    if awConfig.UIBarHeight == nil then
      DEFAULT_CHAT_FRAME:AddMessage("Bar Height: 18", 1, 1, 0.45)
    else
      DEFAULT_CHAT_FRAME:AddMessage("Bar Height: " .. tostring(awConfig.UIBarHeight), 1, 1, 0.45)
    end
    return
  end
  InterfaceOptionsFrame_OpenToCategory(aw.CP.Frame);
  InterfaceOptionsFrame_OpenToCategory(aw.CP.Frame);
end

-- /aw hide
function aw:SlashHide(args)
  if args == nil or args == "" then
    awConfig.HideUI = true
    aw.ui:Hide()
    return
  end
  if args == "title" then
    aw.ui:HideTitle()
    return
  end
  DEFAULT_CHAT_FRAME:AddMessage("/aw hide to hide the window, /aw hide title to hide the title.", 1, 1, 1)
end

-- /aw show
function aw:SlashShow(args)
  if args == nil or args == "" then
    awConfig.HideUI = false
    aw.ui:Show()
    return
  end
  if args == "title" then
    aw.ui:ShowTitle()
    return
  end
  DEFAULT_CHAT_FRAME:AddMessage("/aw show to show the window, /aw show title to show the title.", 1, 1, 1)
end

-- /aw grow
function aw:SlashGrow(args)
  if args == nil then
    DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch: You must specify a direction to grow (up/down)", 1, 1, 1)
    return
  end
  args = strlower(args)
  if args == "up" then
    DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch: Bars will now grow up.", 1, 1, 1)
    if awConfig.GrowDown == true then
      awConfig.GrowDown = false
      aw.ui:ReverseGrowth()
    end
  elseif args == "down" then
    DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch: Bars will now grow down.", 1, 1, 1)
    if awConfig.GrowDown ~= true then
      awConfig.GrowDown = true
      aw.ui:ReverseGrowth()
    end
  else
    DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch: Grow direction can only be 'up' or 'down'", 1, 1, 1)
    return
  end
end

-- /aw lock - Lock or Unlock the Frame
function aw:SlashLock()
  if args == nil then
    if awConfig.LockUI == nil or awConfig.LockUI == true then
      awConfig.LockUI = false
    else 
      awConfig.LockUI = true
    end
  else
    args = strlower(args)
    if args == "on" then
      awConfig.LockUI = true
    elseif args == "off" then
      awConfig.LockUI = false
    else
      DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch: Lock argument must be 'on' or 'off'", 1, 0, 0)
      return
    end
  end
  if awConfig.LockUI == true then
    DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch: Frame is now locked.", 1, 1, 1)
    aw.Frame:SetScript("OnDragStart", aw.Frame.StartMoving)
  else
    DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch: Frame is now locked.", 1, 1, 1)
    aw.Frame:SetScript("OnDragStart", nil)
  end
end

-- /aw on
function aw:SlashOn()
  if awConfig.Enabled == true then
    return
  end
  awConfig.Enabled = true
  DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch is now Enabled.", 1, 1, 1)
end

-- /aw off
function aw:SlashOff()
  if awConfig.Enabled == false then
    return
  end
  awConfig.Enabled = false
  DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch is now Disabled.", 1, 1, 1)
end

-- /aw scale
function aw:SlashScale(args)
  if tonumber(args) == nil then
    DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch: You must specify a numeric value for /aw scale", 1, 1, 1)
    return
  end
  if tonumber(args) < 0.1 or tonumber(args) > 2 then
    DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch: Scale value must be between 0.1 and 2", 1, 1, 1)
    return
  end
  aw.ui:Scale(tonumber(args))
end

-- /aw sort
function aw:SlashSort(args)
  if args == nil then
    DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch: You must specify a direction to sort (asc/desc)", 1, 1, 1)
    return
  end
  args = strlower(args)
  if args == "asc" or args == "ascending" then
    DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch: Sort order set to Ascending", 1, 1, 1)
    if not awConfig.SortAsc then
      awConfig.SortAsc = true
      aw.ui:RefreshUI()
    end
    return
  end
  if args == "desc" or args == "descending" then
    DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch: Sort order set to Descending", 1, 1, 1)
    if awConfig.SortAsc then
      awConfig.SortAsc = false
      aw.ui:RefreshUI()
    end
    return
  end
  DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch: Sort order must be 'asc' or 'desc'", 1, 1, 1)
end

-- /aw width
function aw:SlashWidth(args)
  if args == nil or tonumber(args) == nil then
    DEFAULT_CHAT_FRAME:AddMessage("/aw width requires a numerical argument.", 1, 1, 1)
    return
  end
  local w = tonumber(args)
  if w < 100 or w > 400 then
    DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch: Width must be between 100 and 400", 1, 1, 1)
    return
  end
  if w == awConfig.UIWidth then return end
  awConfig.UIWidth = w
  aw.ui:UpdateWidth()
end

-- /aw height
function aw:SlashHeight(args)
  if args == nil or tonumber(args) == nil then
    DEFAULT_CHAT_FRAME:AddMessage("/aw height requires a numerical argument.", 1, 1, 1)
    return
  end
  local h = tonumber(args)
  if h < 5 or h > 50 then
    DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch: Bar Height must be between 5 and 50", 1, 1, 1)
    return
  end
  DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch: Bar Height is now " .. tostring(h), 1, 1, 1)
  if h == awConfig.UIBarHeight then return end
  awConfig.UIBarHeight = h
  aw.ui:UpdateHeight()
end

-- /aw linger
function aw:SlashLinger(args)
  if args == nil or tonumber(args) == nil then
    DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch: Linger requires numerical argument.", 1, 0, 0)
  end
  awConfig.Linger = tonumber(args)
  DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch: Linger set to " .. tostring(awConfig.Linger) .. " seconds.", 1, 1, 1)
end

-- /aw debug
function aw:SlashDebug(args)
  -- Finished with this but I'll leave it in coz I'm too lazy to redo it
  -- if I need it later.
  for i = #aw.Data.Units, 1, -1 do
    local u = aw.Data.Units[i]
    DEFAULT_CHAT_FRAME:AddMessage("[" .. u.UnitName .. "]: " .. u.BuffExpiry, 1, 1, 1)
  end
  DEFAULT_CHAT_FRAME:AddMessage("Units: " .. tostring(#aw.Data.Units) .. " Bars: " .. tostring(#aw.ui.awBars), 1, 1, 1)
end

-------------------
-- Event Handlers -
-------------------
-- HeartBeat.
function aw:HeartBeat(elapsed)
  if not aw:IsEnabled() then return end
  if aw.Data.LastUpdate + 0.1 > GetTime() then return end
  aw.DataUpdateAll()
end

-- Set the Spec (0-2 for Priests, 3 for any other class)
function aw:UpdatePlayerSpec()
  local nClassID = 0
  _,_,nClassID = UnitClass("PLAYER")
  if nClassID == 5 then
    aw.Data.Player.Spec = GetSpecialization()
  else
    aw.Data.Player.Spec = 4
  end
end

-- Is the addon currently enabled?
function aw:IsEnabled()
  if awConfig.Enabled == false then
    return false
  elseif aw.Data.Player.Spec ~= 1 then
    return false
  end
  return true
end

-- Handler for PLAYER_LOGIN
function aw:playerLogin()
  aw.Data.Player.GUID = UnitGUID("PLAYER")
  aw.Data.Player.Name = UnitName("PLAYER")
  aw:UpdatePlayerSpec()
  aw.Frame:SetScript("OnUpdate", aw.HeartBeat)
end

-- Handler for PLAYER_ENTERING_WORLD
function aw:playerEnteringWorld()
  -- Set some defaults in case
  if awConfig.Linger == nil then awConfig.Linger = 2 end
  if awConfig.Enabled == nil then awConfig.Enabled = true end
  if awConfig.Debug == nil then awConfig.Debug = false end
  if awConfig.GrowDown == nil then awConfig.GrowDown = false end
  if awConfig.SortAsc == nil then awConfig.SortAsc = false end
  if awConfig.Scale == nil then awConfig.Scale = 1 end
  if awConfig.LockUI == nil then awConfig.LockUI = false end
  if awConfig.HideUI == nil then awConfig.HideUI = false end
  if awConfig.UIWidth == nil then awConfig.UIWidth = 200 end
  if awConfig.UIBarHeight == nil then awConfig.UIBarHeight = 18 end
  if awConfig.UIHideTitle == nil then awConfig.UIHideTitle = false end
  if awConfig.UIHideTitle then
    aw.ui:HideTitle()
  else
    aw.ui:ShowTitle()
  end
  if aw:IsEnabled() and not awConfig.HideUI then
    aw.ui:Show()
  else
    aw.ui:Hide()
  end
end

-- Handler for ADDON_LOADED
function aw:addonLoaded()
  local sEn = "Enabled"
  if awConfig.Enabled == false then
    sEn = "Disabled"
  end
  if not aw.Data.UILoaded then
    aw.ui:BuildFrame()
    aw.CP:BuildFrame()
    aw.ui:Hide()
    aw.Data.UILoaded = true
  end
  if awConfig.LockUI then
    aw:SlashLock("on")
  end
  if awConfig.UIPos ~= nil then
    aw.Frame:SetPoint(awConfig.UIPos.point, awConfig.UIPos.relTo, awConfig.UIPos.relPoint, awConfig.UIPos.XOff, awConfig.UIPos.YOff)
  end
  DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch v" .. tostring(aw.Data.Version) .. " Loaded (/aw) - " .. sEn, 1, 1, 0.45)
end

-- Handler for COMBAT_LOG_EVENT_UNFILTERED
function aw:CombatLogHandler()
  local timestamp,Event,_,SGUID,_,_,_,TGUID,TName,_,_,SpellID,SpellName = CombatLogGetCurrentEventInfo()
  if SGUID ~= aw.Data.Player.GUID then return end
  if Event == "SPELL_AURA_APPLIED" then
    -- Add the unit if we applied atonement.
    if SpellID == awSpellIDs.Atonement then
      aw:DataAddUnit(TGUID, TName, GetTime() + aw.Data.NextDuration)
      return
    end
  elseif Event == "SPELL_AURA_REMOVED" then
    -- Update the unit if atonement was removed.
    -- Handle: Expired, Purged, Self-Removed (Why you do this?)
    -- Consider: Linger
    if SpellID == awSpellIDs.Atonement then
      aw:DataSetGUIDExpiry(TGUID, GetTime())
      aw:DataUpdateGUID(TGUID)
      return
    end
  elseif Event == "SPELL_AURA_REFRESH" then
    -- Update the unit if atonement was refreshed.
    if SpellID == awSpellIDs.Atonement then
      if aw.Data.LastSpellID == awSpellIDs.Evangelism then return end
      aw:DataAddUnit(TGUID, TName, GetTime() + aw.Data.NextDuration)
      -- aw:DataUpdateGUID(TGUID)
      return
    end
  elseif Event == "SPELL_CAST_SUCCESS" then
    -- Set the Last Spell ID and Next Duration for casts of ShadowMend, PW:Shield, PW:Radiance
    -- Update The Expiration for casts of Evangelism
    -- TODO-02: Azerite Checks (Enduring Luminescence, Moment of Repose, etc.)
    if SpellID == awSpellIDs.Evangelism then
      aw:DataUpdateAllExpiry(awDurations.Evangelism)
    elseif SpellID == awSpellIDs.ShadowMend then
      aw.Data.LastSpellID = awSpellIDs.ShadowMend
      aw.Data.NextDuration = awDurations.ShadowMend
    elseif SpellID == awSpellIDs.Shield then
      aw.Data.LastSpellID = awSpellIDs.Shield
      aw.Data.NextDuration = awDurations.Shield
    elseif SpellID == awSpellIDs.Radiance then
      aw.Data.LastSpellID = awSpellIDs.Radiance
      aw.Data.NextDuration = awDurations.Radiance
    end
  end
  if SpellID == awSpellIDs.Atonement then
    DEFAULT_CHAT_FRAME:AddMessage("AtoneWatch: Unhandled " .. SpellName .. " Event: " .. Event, 1,1,1)
  end
  aw:HeartBeat(0)
end

-- Handler for all events.
function OnEvent(self, event, ... )
  local parm = {...}
  if event == "PLAYER_LOGIN" then
    aw:playerLogin()
  elseif event == "ADDON_LOADED" and parm[1] == "AtoneWatch" then
    aw:addonLoaded()
  elseif event == "COMBAT_LOG_EVENT_UNFILTERED" then
    aw:CombatLogHandler(...)
  elseif event == "PLAYER_ENTERING_WORLD" then
    aw:playerEnteringWorld()
  elseif event == "ACTIVE_TALENT_GROUP_CHANGED" then
    aw:UpdatePlayerSpec()
    if aw:IsEnabled() and not awConfig.HideUI then
      aw.ui:Show()
    else 
      aw.ui:Hide()
    end
  end
end

--------------------
-- Data Functions --
--------------------
-- Sort the list.
function aw:DataSort()
  if #aw.Data.Units < 2 then return end
  if (awConfig.SortAsc) then
    table.sort(aw.Data.Units, function(a,b) if a.BuffRem == b.BuffRem then return a.UnitName < b.UnitName else return a.BuffRem < b.BuffRem end end)
  else
    table.sort(aw.Data.Units, function(a,b) if a.BuffRem == b.BuffRem then return a.UnitName < b.UnitName else return a.BuffRem > b.BuffRem end end)
  end
end

-- Get an entry in the list.
function aw:DataGetUnit(GUID)
  local i = 1
  for i = 1, #aw.Data.Units, 1 do
    if aw.Data.Units[i].UnitGUID == GUID then
      return i
    end
  end
  return nil
end

-- Add an entry to the list.
function aw:DataAddUnit(GUID, Name, Expiry)
  local ExistIndex = aw:DataGetUnit(GUID)
  local tNow = GetTime()
  if Expiry - tNow > aw.Data.MaxDuration then
    aw.Data.MaxDuration = Expiry - tNow
  end
  if ExistIndex ~= nil then
    if aw.Data.Units[ExistIndex].BuffExpiry < Expiry then
      aw.Data.Units[ExistIndex].BuffApplied = tNow
      aw.Data.Units[ExistIndex].BuffExpiry = Expiry
      aw.Data.Units[ExistIndex].BuffRem = tNow - Expiry
    end
  else
    local UnitID = aw:Find_UnitID(GUID)
    local lUnitClass = ""
    if UnitID ~= "" then lUnitClass = select(2, UnitClass(UnitID)) end
    table.insert(aw.Data.Units, {UnitGUID = GUID, UnitName = Name, BuffApplied = tNow, BuffExpiry = Expiry, UnitUnitID = UnitID, UnitClass = lUnitClass, BuffRem = Expiry - tNow})
  end
  aw:DataSort()
  aw:DataUpdateGUID(GUID)
end

-- Remove an entry from the list.
function aw:DataDelUnit(GUID)
  local i = aw:DataGetUnit(GUID)
  if i ~= nil then 
    table.remove(aw.Data.Units, i) 
  end
  aw.Data.MaxDuration = 0
  for i = #aw.Data.Units, 1, -1 do
    if aw.Data.Units[i].BuffExpiry - aw.Data.Units[i].BuffApplied > aw.Data.MaxDuration then
      aw.Data.MaxDuration = aw.Data.Units[i].BuffExpiry - aw.Data.Units[i].BuffApplied
    end
  end
  aw:DataSort()
  aw.ui:RefreshUI()
end

-- Update a single unit (by GUID)
function aw:DataUpdateGUID(GUID)
  local i = aw:DataGetUnit(GUID)
  if i ~= nil and i > 0 then
    aw:DataUpdateID(i)
    aw.ui:RefreshUI()
  end
end

-- Update a single unit (by Index)
function aw:DataUpdateID(nID)
  if nID == nil or nID < 1 or nID > #aw.Data.Units then return end
  local GUID = aw.Data.Units[nID].UnitGUID
  local tNow = GetTime()
  -- Make sure we have a Unit ID, else we can't really do much.
  if GUID ~= UnitGUID(aw.Data.Units[nID].UnitUnitID) then
    aw.Data.Units[nID].UnitUnitID = aw:Find_UnitID(GUID)
  end
  if aw.Data.Units[nID].UnitUnitID == "" then
    aw.Data.Units[nID].HealthPct = -2
    aw.Data.Units[nID].BuffRem = aw.Data.Units[nID].BuffExpiry - tNow
    return
  end
  -- If the units class was not set, set it.
  if aw.Data.Units[nID].UnitClass == "" then
    aw.Data.Units[nID].UnitClass = select(2, UnitClass(aw.Data.Units[nID].UnitUnitID))
  end
  -- If the unit is dead don't calculate their health.
  if UnitIsDead(aw.Data.Units[nID].UnitUnitID) then
    aw.Data.Units[nID].HealthPct = -1
    if aw.Data.Units[nID].BuffExpiry > tNow then
      aw.Data.Units[nID].BuffExpiry = tNow
      aw.Data.Units[nID].BuffRem = aw.Data.Units[nID].BuffExpiry - tNow
    end
    return
  end
  -- Buff Expiry
  local i = 1
  for i = 1, 40, 1 do
    local expirationTime = 0
    local caster = 0
    local spellID = 0
    _,_,_,_,_, expirationTime, caster,_,_, spellID, _,_, isCastByPlayer, _,_ = UnitAura(aw.Data.Units[nID].UnitUnitID, i)
    if spellID == awSpellIDs.Atonement and caster == "player" then
      aw.Data.Units[nID].BuffExpiry = expirationTime
      break
    end
  end
  -- Buff Remaining
  aw.Data.Units[nID].BuffRem = aw.Data.Units[nID].BuffExpiry - tNow
  -- Health Percentage
  local nMaxHP = UnitHealthMax(aw.Data.Units[nID].UnitUnitID)
  if nMaxHP < 1 then
    aw.Data.Units[nID].HealthPct = "-1"
    return
  end
  aw.Data.Units[nID].HealthPct = 100 * UnitHealth(aw.Data.Units[nID].UnitUnitID) / UnitHealthMax(aw.Data.Units[nID].UnitUnitID)
end

-- Updates all entries in the list
function aw:DataUpdateAll()
  -- Let's not be too spammy.
  if aw.Data.LastUpdate + 0.1 > GetTime() then return end
  aw.Data.LastUpdate = GetTime()
  if aw.Data.Units == nil or #aw.Data.Units < 1 then return end
  local i = 1
  for i = #aw.Data.Units, 1, -1 do
    if aw.Data.Units[i].BuffExpiry + awConfig.Linger < GetTime() then
      table.remove(aw.Data.Units, i) 
      -- Fix an issue where sometimes the UI would not update on unit removal
      -- because the UI had been updated too recently.
      aw.ui.LastUpdate = 0
    else
      aw:DataUpdateID(i)
    end
  end
  aw:DataSort()
  aw.ui:RefreshUI()
end

-- Update the expiration of a specific GUID.
function aw:DataSetGUIDExpiry(GUID, tExpiry)
  local i = aw:DataGetUnit(GUID)
  if i == nil then return end
  aw.Data.Units[i].BuffExpiry = tExpiry
  aw.Data.Units[i].BuffRem = tExpiry - GetTime()
  aw:DataSort()
end

-- Update all expirations in the list (Evangelism)
function aw:DataUpdateAllExpiry(variance)
  for i = 1, #aw.Data.Units, 1 do
    if aw.Data.Units[i].BuffExpiry > GetTime() then
      aw.Data.Units[i].BuffApplied = aw.Data.Units[i].BuffApplied + variance
      aw.Data.Units[i].BuffExpiry = aw.Data.Units[i].BuffExpiry + variance
    end
  end
end

------------------------
-- Scanning Functions --
------------------------
-- Find a GUID in Nameplates
function aw:Find_UnitInNameplates(GUID)
  local nCnt = 1
  while UnitExists("nameplate" .. tostring(nCnt)) do
    if UnitGUID("nameplate" .. tostring(nCnt)) == GUID then
      return "nameplate" .. tostring(nCnt)
    end
    if UnitExists("nameplate" .. tostring(nCnt) .. "target") then
      if UnitGUID("nameplate" .. tostring(nCnt) .. "target") == GUID then
	return "nameplate" .. tostring(nCnt) .. "target"
      end
    end
    nCnt = nCnt + 1
  end
  return ""
end

-- Find a GUID in Group(Party/Raid) Members or Pets
function aw:Find_UnitInGMP(GUID, suffix)
  local gType = LE_PARTY_CATEGORY_HOME
  if IsInGroup(LE_PARTY_CATEGORY_INSTANCE) then
    gType = LE_PARTY_CATEGORY_INSTANCE
  end
  local nGroupSize = GetNumGroupMembers(gType)
  if nGroupSize < 1 then return "" end
  local sPrefix = "party" .. suffix
  if IsInRaid() then sPrefix = "raid" .. suffix end
  local nCnt = 0
  for nCnt = 1, nGroupSize, 1 do
    if UnitExists(sPrefix .. tostring(nCnt)) then
      if UnitGUID(sPrefix .. tostring(nCnt)) == GUID then
	return sPrefix .. tostring(nCnt)
      end
    end
  end
  return ""
end
  
-- Find a GUID in Party/Raid
function aw:Find_UnitInGroup(GUID)
  return aw:Find_UnitInGMP(GUID, "")
end

-- Find a GUID in Party/Raid Pets
function aw:Find_UnitInGroupPets(GUID)
  return aw:Find_UnitInGMP(GUID, "pet")
end

-- Find a GUID in the list of Boss Targets
function aw:Find_UnitInBossTargets(GUID)
  local nCnt = 1
  while UnitExists("boss" .. tostring(nCnt)) do
    if UnitExists("boss" .. tostring(nCnt) .. "target") then
      if UnitGUID("boss" .. tostring(nCnt) .. "target") == GUID then
	return "boss" .. tostring(nCnt) .. "target"
      end
    end
    nCnt = nCnt + 1
  end
  return ""
end

-- Find a valid UnitID for a GUID.
function aw:Find_UnitID(GUID)
  -- Is it the current player?
  if GUID == aw.Data.Player.GUID then
    return "PLAYER"
  end
  -- Are they our Target/Target's Target
  if UnitExists("target") then
    if UnitGUID("target") == GUID then
      return "target"
    end
    if UnitExists("targettarget") then
      if UnitGUID("targettarget") == GUID then
	return "targettarget"
      end
    end
  end
  -- Are they our focus target?
  if UnitExists("focus") then
    if UnitGUID("focus") == GUID then return "focus" end
    if UnitExists("focustarget") then
      if UnitGUID("focustarget") == GUID then return "focustarget" end
    end
  end
  -- Are they the mouseover target?
  if UnitExists("mouseover") then
    if UnitGUID("mouseover") == GUID then return "mouseover" end
  end
  -- Are they in our Party/Raid (Instance Group if available else Home Group)?
  local RetVal = ""
  RetVal = aw:Find_UnitInGroup(GUID)
  if RetVal ~= "" then return RetVal end
  -- Are they being targeted by a boss?
  RetVal = aw:Find_UnitInBossTargets(GUID)
  if RetVal ~= "" then return RetVal end
  -- Are they a party/raid pet?
  RetVal = aw:Find_UnitInGroupPets(GUID)
  if RetVal ~= "" then return RetVal end
  -- Are they my pet (why am I even checking this tbh)
  if UnitExists("pet") then
    if UnitGUID("pet") == GUID then return "pet" end
  end
  -- Are they a nameplate or nameplate target?
  RetVal = aw:Find_UnitInNameplates(GUID)
  if RetVal ~= "" then return RetVal end
  return ""
end

aw.Frame:SetScript("OnEvent", OnEvent)

--------------------
-- User Interface --
--------------------
local ui = {
  awBars = {},
  LastUpdate = 0
}
aw.ui = ui

-- Reverse the growth when bars exist already.
function ui:ReverseGrowth()
  awConfig.UIPos.point, awConfig.UIPos.relTo, awConfig.UIPos.relPoint, awConfig.UIPos.XOff, awConfig.UIPos.YOff = aw.Frame:GetPoint(1)
  aw.Frame:Hide()
  if awConfig.GrowDown then
    if awConfig.UIHideTitle and #aw.ui.awBars > 0 then
      aw.Frame:SetPoint(awConfig.UIPos.point, awConfig.UIPos.relTo, awConfig.UIPos.relPoint, awConfig.UIPos.XOff, awConfig.UIPos.YOff - 16 + (#aw.ui.awBars * awConfig.UIBarHeight))
    else
      aw.Frame:SetPoint(awConfig.UIPos.point, awConfig.UIPos.relTo, awConfig.UIPos.relPoint, awConfig.UIPos.XOff, awConfig.UIPos.YOff + (#aw.ui.awBars * awConfig.UIBarHeight))
    end
  else
    if awConfig.UIHideTitle and #aw.ui.awBars > 0 then
      aw.Frame:SetPoint(awConfig.UIPos.point, awConfig.UIPos.relTo, awConfig.UIPos.relPoint, awConfig.UIPos.XOff, awConfig.UIPos.YOff + 16 - (#aw.ui.awBars * awConfig.UIBarHeight))
    else
      aw.Frame:SetPoint(awConfig.UIPos.point, awConfig.UIPos.relTo, awConfig.UIPos.relPoint, awConfig.UIPos.XOff, awConfig.UIPos.YOff - (#aw.ui.awBars * awConfig.UIBarHeight))
    end
  end
  aw.ui:UpdateHeight()
  awConfig.UIPos.point, awConfig.UIPos.relTo, awConfig.UIPos.relPoint, awConfig.UIPos.XOff, awConfig.UIPos.YOff = aw.Frame:GetPoint(1)
end

-- Show the Title
function ui:ShowTitle()
  awConfig.UIHideTitle = false
  aw.Frame:SetBackdrop({
    bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize=16,
    edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize=2,
    insets = { left = 0, right = 0, top = 0, bottom = 0, },
  })
  aw.Frame:SetBackdropColor(0.75,0.75,0.75,0.5)
  aw.Frame.titleText:Show()
  if not awConfig.LockUI then
    aw.Frame:SetScript("OnDragStart", aw.Frame.StartMoving)
  end
  aw.ui:UpdateHeight()
end

-- Hide the Title
function ui:HideTitle()
  awConfig.UIHideTitle = true
  aw.Frame:SetBackdrop(nil)
  aw.Frame:SetBackdropColor(0.75,0.75,0.75,0)
  aw.Frame:SetScript("OnDragStart", nil)
  aw.Frame.titleText:Hide()
  aw.ui:UpdateHeight()
end

-- Hide the UI
function ui:Hide()
  if aw.Frame then
    aw.Frame:Hide()
  end
end

-- Show the UI
function ui:Show()
  if not aw:IsEnabled() then
    aw.Frame:Hide()
    return
  end
  if aw.Frame then
    aw.Frame:Show()
  end
end

-- Update the Height
function ui:UpdateHeight()
  if awConfig.UIBarHeight == nil then return end
  aw.Frame:Hide()
  -- Set up the Initial Bar
  if awConfig.GrowDown then
    local hTitle = 0 - 17
    if awConfig.UIHideTitle then hTitle = 0 end
    for i = 1, #aw.ui.awBars, 1 do
      yOffset = hTitle - ((i-1) * awConfig.UIBarHeight)
      aw.ui.awBars[i]:SetPoint("TOPLEFT", 2, yOffset)
      aw.ui.awBars[i]:SetHeight(awConfig.UIBarHeight)
    end
  else
    local hTitle = 1
    if awConfig.UIHideTitle then hTitle = 0 - 16 end
    for i = 1, #aw.ui.awBars, 1 do
      yOffset = i * awConfig.UIBarHeight + hTitle
      aw.ui.awBars[i]:SetPoint("TOPLEFT", 2, yOffset)
      aw.ui.awBars[i]:SetHeight(awConfig.UIBarHeight)
    end
  end
  if aw:IsEnabled() and not awConfig.HideUI then aw.Frame:Show() end
end

-- Update the Width
function ui:UpdateWidth()
  if awConfig.UIWidth == nil then return end
  aw.Frame:SetWidth(awConfig.UIWidth)
  for i = 1, #ui.awBars, 1 do
    ui.awBars[i]:SetWidth(awConfig.UIWidth)
  end
end

-- Scale the UI
function ui:Scale(nVal)
  nVal = tonumber(nVal)
  if not nVal or nVal < 0.1 or nVal > 2 then
    nVal = 1
  else
    awConfig.Scale = nVal
  end
  aw.Frame:SetScale(nVal)
end

-- Add a new Bar
function ui:AddBar()
  local newbar = CreateFrame("StatusBar", nil, aw.Frame)
  local yOffset = 0
  if awConfig.GrowDown then
    if awConfig.UIHideTitle then
      yOffset = 0 - (#aw.ui.awBars * (awConfig.UIBarHeight))
    else
      yOffset = 0 - (#aw.ui.awBars * (awConfig.UIBarHeight)) - 17
    end
  else
    if awConfig.UIHideTitle then
      yOffset = ((#aw.ui.awBars + 1) * awConfig.UIBarHeight) - 17
    else
      yOffset = ((#aw.ui.awBars + 1) * awConfig.UIBarHeight)
    end
  end
  -- Set up the Initial Bar
  newbar:SetPoint("TOPLEFT", 2, yOffset)
  newbar:SetWidth(awConfig.UIWidth)
  newbar:SetHeight(awConfig.UIBarHeight)
  newbar:SetStatusBarTexture("Interface\\TargetingFrame\\UI-StatusBar")
  newbar:GetStatusBarTexture():SetHorizTile(false)
  newbar:GetStatusBarTexture():SetVertTile(false)
  newbar:SetStatusBarColor(0.9, 0.9, 0.9)
  newbar:SetMinMaxValues(0, 100)
  newbar:SetValue(100)
  newbar:EnableMouse(true)
  newbar:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
                      edgeFile = nil,
		      tile = true, tileSize=16, edgeSize = 0,
		      insets = { 0, 0, 0, 0 }
		    })
  newbar:SetBackdropColor(0,0,0,0.25)
  newbar:Hide()
  -- Add Name Text
  newbar.UnitName = newbar:CreateFontString(nil, "OVERLAY")
  newbar.UnitName:SetFont("Fonts\\ARIALN.TTF", 12, "OUTLINE")
  newbar.UnitName:SetJustifyH("LEFT")
  newbar.UnitName:SetShadowOffset(1, -1)
  newbar.UnitName:SetTextColor(1, 1, 1)
  newbar.UnitName:SetText(UnitName("PLAYER"))
  newbar.UnitName:SetPoint("LEFT", newbar, "LEFT", 2, 1)
  -- Add Time Text
  newbar.Duration = newbar:CreateFontString(nil, "OVERLAY")
  newbar.Duration:SetFont("Fonts\\ARIALN.TTF", 10, "OUTLINE")
  newbar.Duration:SetJustifyH("RIGHT")
  newbar.Duration:SetShadowOffset(1, -1)
  newbar.Duration:SetTextColor(0, 0.9, 0)
  newbar.Duration:SetText("15.0s")
  newbar.Duration:SetPoint("RIGHT", newbar, "RIGHT", 1, 1)
  -- Add Health Text
  newbar.Health = newbar:CreateFontString(nil, "OVERLAY")
  newbar.Health:SetFont("Fonts\\ARIALN.TTF", 12, "OUTLINE")
  newbar.Health:SetJustifyH("RIGHT")
  newbar.Health:SetShadowOffset(1, -1)
  newbar.Health:SetTextColor(0, 0.9, 0)
  newbar.Health:SetText("100%")
  newbar.Health:SetPoint("RIGHT", newbar, "RIGHT", -30, 1)
  newbar:Show()
  newbar.awUnitGUID = ""
  table.insert(aw.ui.awBars, newbar)
end

-- Refresh the UI
function ui:RefreshUI()
  -- Let's not be too spammy.
  if not aw:IsEnabled() then return end
  if awConfig.HideUI then return end
  if aw.ui.LastUpdate + 0.05 > GetTime() then return end
  aw.ui.LastUpdate = GetTime()
  for i = 1, #aw.Data.Units, 1 do
    local awUnit = aw.Data.Units[i]
    if i > #aw.ui.awBars then
      ui:AddBar()
    end
    -- Set Name
    if aw.ui.awBars[i].awUnitGUID ~= awUnit.UnitGUID then
      aw.ui.awBars[i].UnitName:SetText(awUnit.UnitName)
    end
    aw.ui.awBars[i].awUnitGUID = awUnit.UnitGUID
    local cR, cG, cB = 0.5, 0.5, 0.5 
    if awUnit.UnitClass ~= "" then
      cR, cG, cB, _ = GetClassColor(awUnit.UnitClass)
    end
    aw.ui.awBars[i].UnitName:SetTextColor(cR, cG, cB)
    if cR <= 0.15 then cR = 0 else cR = cR - 0.15 end
    if cG <= 0.15 then cG = 0 else cG = cG - 0.15 end
    if cB <= 0.15 then cB = 0 else cB = cB - 0.15 end
    aw.ui.awBars[i]:SetStatusBarColor(cR, cG, cB)
    -- Health
    if awUnit.HealthPct == -2 then
      aw.ui.awBars[i].Health:SetText("??")
      aw.ui.awBars[i].Health:SetTextColor(0.75, 0.75, 0.75)
    elseif awUnit.HealthPct == -1 then
      aw.ui.awBars[i].Health:SetText("RIP")
      aw.ui.awBars[i].Health:SetTextColor(0.7, 0.7, 0.7)
    else
      if awUnit.HealthPct < 30 then
	aw.ui.awBars[i].Health:SetTextColor(0.8, 0, 0)
	if awUnit.HealthPct < 10 then
	  aw.ui.awBars[i].Health:SetText(format("%.1f%%", awUnit.HealthPct))
	else
	  aw.ui.awBars[i].Health:SetText(floor(awUnit.HealthPct)..'%')
	end
      elseif aw.Data.Units[i].HealthPct < 75 then
	aw.ui.awBars[i].Health:SetTextColor(0.9,0.9,0)
	aw.ui.awBars[i].Health:SetText(floor(awUnit.HealthPct)..'%')
      else
	aw.ui.awBars[i].Health:SetTextColor(0,0.9,0)
	aw.ui.awBars[i].Health:SetText(floor(awUnit.HealthPct)..'%')
      end
    end
    -- Duration
    if (awUnit.BuffRem <= 0) then 
      aw.ui.awBars[i].Duration:SetText("exp")
      aw.ui.awBars[i].Duration:SetTextColor(0.7,0,0)
      aw.ui.awBars[i]:SetValue(0)
    else
      aw.ui.awBars[i].Duration:SetText(format("%.1fs", awUnit.BuffRem))
      aw.ui.awBars[i].Duration:SetTextColor(0,0.9,0)
      aw.ui.awBars[i]:SetValue(awUnit.BuffRem)
    end
    aw.ui.awBars[i]:SetMinMaxValues(0, aw.Data.MaxDuration)
    aw.ui.awBars[i]:Show()
  end
  for i = #aw.ui.awBars, #aw.Data.Units + 1, -1 do
    aw.ui.awBars[i]:Hide()
  end
end

-- Create the UI Frame
function ui:BuildFrame()
  local f = aw.Frame
  local fwidth = 200
  if awConfig and awConfig.UIWidth then
    fwidth = awConfig.UIWidth
  end
  if fwidth == nil then
    fwidth = 200
    if awConfig then
      awConfig.UIWidth = 200
    end
  end
  f:SetSize(fwidth,16)
  f:SetPoint("CENTER", UIParent)
  f:SetBackdrop({
    bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize=16,
    edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize=2,
    insets = { left = 0, right = 0, top = 0, bottom = 0, },
  })
  f:SetBackdropColor(0.75,0.75,0.75,0.5)
  f:SetMovable(true)
  f:EnableMouse(true)
  f:SetClampedToScreen(true)
  f:RegisterForDrag("LeftButton")
  f:SetUserPlaced(true)
  f:SetScript("OnDragStart", f.StartMoving)
  f:SetScript("OnDragStop", function()
    f:StopMovingOrSizing()
    if awConfig.UIPos == nil then awConfig.UIPos = {} end
    awConfig.UIPos.point, awConfig.UIPos.relTo, awConfig.UIPos.relPoint, awConfig.UIPos.XOff, awConfig.UIPos.YOff = f:GetPoint(1)
  end)
  f:Hide()

  f.titleText = f:CreateFontString(nil, "OVERLAY")
  f.titleText:SetPoint("CENTER", f, "CENTER", 1, 1)
  f.titleText:SetFont("Fonts\\FRIZQT__.TTF", 10, "OUTLINE")
  f.titleText:SetJustifyH("LEFT")
  f.titleText:SetShadowOffset(1, -1)
  f.titleText:SetTextColor(1, 1, 0.45)
  f.titleText:SetText( "Atonements" )
end
